#pragma once

#include <cstddef>
#include <cstdlib>
#include <limits>
#include <new>
#include <vector>
#include <algorithm>
#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/voxel_painter.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/sprite.hpp>
#include "color.hpp"
#include "matrix.hpp"
#include "shapes/rectangle.hpp"

namespace BGTU::ComputerGraphicsLabWork::Utilities {
	class DefaultVoxelDrawerCache {
	private:
		class ZElement {
		public:
			ZElement *next;
			double const z;
			RendererApi::Color::Transparent const color;
			RendererApi::SpriteMetadata *const owner;

			ZElement(ZElement *next, double z, RendererApi::Color::Transparent color, RendererApi::SpriteMetadata *owner) : next{next}, z{z}, color{color}, owner{owner} {}

			void link_after(ZElement *new_next) {
				new_next->next = this->next;
				this->next = new_next;
			}
		};

		class ZElementAllocationBuffer {
		private:
			union Cell {
				ZElement allocated;
				struct Empty {
					Cell *next_empty;
				} empty;
			};


			class Page {
			public:
				Page *next;
			private:

				std::size_t initialized;
				std::size_t capacity;

				Cell buffer[0];

			public:
				explicit Page(Page *next, std::size_t capacity) : next{next}, initialized{0}, capacity{capacity}, buffer{} {};

				Cell *try_allocate();

				static std::size_t calc_capacity(std::size_t bytes) {
					return (bytes - offsetof(Page, buffer)) / sizeof(Cell);
				};

				void reset_allocations() {
					this->initialized = 0;
				}
			};

			Cell *next_unallocated;
			Page *first_allocated_page;
			Page *last_allocated_page;
			Page *next_unfinished_page;
			Page sentinel;

		public:
			ZElementAllocationBuffer() :
					next_unallocated{nullptr}, sentinel{nullptr, 0},
					first_allocated_page{&this->sentinel}, last_allocated_page{this->first_allocated_page}, next_unfinished_page{this->first_allocated_page} {};

		private:
			Page *_native_extend();

		public:
			template<class ...args_t>
			ZElement *alloc_elem(args_t...args);

			void free_elem(ZElement *e);

		private:
			void _native_free_page(Page *p);

		public:
			void release_resources();

			void reset_allocations();

			~ZElementAllocationBuffer();
		};

		struct pixel_trace_metadata {
		public:
			pixel_trace_metadata() : nearest_z{std::numeric_limits<double>::max()}, voxels{nullptr} {};

			double nearest_z;
			ZElement *voxels;
		};

	public:
		class VoxelPainterImpl : public RendererApi::VirtualVoxelPainter {
			friend
			class DefaultVoxelDrawerCache;

		private:
			RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t _width;
			RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t _height;

			pixel_trace_metadata *pixels_metadata;

			ZElementAllocationBuffer *pixel_trace_elements_allocator;
			RendererApi::SpriteMetadata *current_artist;

			VoxelPainterImpl(RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t width, RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t height, pixel_trace_metadata *pixels_metadata, ZElementAllocationBuffer *pixel_trace_elements_allocator);

		public:
			[[nodiscard]] inline RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t width() const final {
				return this->_width;
			}

			[[nodiscard]] inline RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t height() const final {
				return this->_height;
			}

		private:
			inline pixel_trace_metadata *at(std::uint_fast16_t x, std::uint_fast16_t y) {
				return &(this->pixels_metadata[y * this->_width + x]);
			}

			void _add_voxel(long long x, long long y, double z, RendererApi::Color::Transparent c, RendererApi::SpriteMetadata *owner);

		public:
			inline void add_voxel(long long x, long long y, double z, RendererApi::Color::Transparent c) final {
				return this->add_voxel(x, y, z, c, this->current_artist);
			}

			void add_voxel(long long x, long long y, double z, RendererApi::Color::Transparent c, RendererApi::SpriteMetadata *owner) final;


			void fill_rectangle(
					RendererApi::PointI<2>::component_t left,
					RendererApi::PointI<2>::component_t top,
					RendererApi::PointI<2>::component_t right,
					RendererApi::PointI<2>::component_t bottom,
					double z,
					RendererApi::Color::Transparent fill,
					RendererApi::SpriteMetadata *owner
			);

			inline void fill_rectangle(
					RendererApi::PointI<2>::component_t left,
					RendererApi::PointI<2>::component_t top,
					RendererApi::PointI<2>::component_t right,
					RendererApi::PointI<2>::component_t bottom,
					double z,
					RendererApi::Color::Transparent fill
			) {
				this->fill_rectangle(left, top, right, bottom, z, fill, this->current_artist);
			}

			void draw_h_line(
					RendererApi::PointI<2>::component_t y,
					RendererApi::PointI<2>::component_t left,
					RendererApi::PointI<2>::component_t right,
					double z,
					RendererApi::Color::Transparent color,
					RendererApi::SpriteMetadata *owner
			);

			inline void draw_h_line(
					RendererApi::PointI<2>::component_t y,
					RendererApi::PointI<2>::component_t left,
					RendererApi::PointI<2>::component_t right,
					double z,
					RendererApi::Color::Transparent color
			) {
				this->draw_h_line(y, left, right, z, color, this->current_artist);
			}
		};

	private:
		struct {
			pixel_trace_metadata *buffer;
			std::size_t size;
		} pixel_metadata_cache;
		ZElementAllocationBuffer pixel_trace_elements_allocator;

	public:
		DefaultVoxelDrawerCache();

	private:
		static ZElement *sort_zelements_desc(DefaultVoxelDrawerCache::ZElement *start) noexcept;

	public:
		template<class sprite_data_t, class voxel_exporter_t = RendererApi::VirtualVoxelPainter::Exporter>
		void render(
				RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t w,
				RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t h,
				voxel_exporter_t *frame,
				RendererApi::Color bg,
				RendererApi::Sprite<sprite_data_t, VoxelPainterImpl> *const *sprites,
				std::size_t sprites_count,
				sprite_data_t const *sprite_data
		);

		~DefaultVoxelDrawerCache();
	};

	template<class... args_t>
	DefaultVoxelDrawerCache::ZElement *DefaultVoxelDrawerCache::ZElementAllocationBuffer::alloc_elem(args_t... args) {
		auto cell = this->next_unallocated;
		if (cell != nullptr) {
			this->next_unallocated = cell->empty.next_empty;
			return new(&(cell->allocated)) ZElement{args...};
		}

		while (this->next_unfinished_page != nullptr && (cell = this->next_unfinished_page->try_allocate()) == nullptr) {
			this->next_unfinished_page = this->next_unfinished_page->next;
		}

		if (cell == nullptr) {
			while ((cell = this->last_allocated_page->try_allocate()) == nullptr)
				this->_native_extend();
			this->next_unfinished_page = this->last_allocated_page;
		}
		return new(&(cell->allocated)) ZElement{args...};
	}


	template<class sprite_data_t, class voxel_exporter_t>
	void DefaultVoxelDrawerCache::render(
			RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t w,
			RendererApi::VirtualVoxelPainter::visible_pixel_coordinate_fast_t h,
			voxel_exporter_t *frame,
			RendererApi::Color bg,
			RendererApi::Sprite<sprite_data_t, VoxelPainterImpl> *const *sprites,
			std::size_t sprites_count,
			sprite_data_t const *sprite_data
	) {
		if (w == 0 || h == 0)
			return;

		if (this->pixel_metadata_cache.buffer == nullptr || this->pixel_metadata_cache.size < w * h) {
			this->pixel_metadata_cache.buffer = static_cast<pixel_trace_metadata *>(std::realloc(this->pixel_metadata_cache.buffer, w * h * sizeof(pixel_trace_metadata)));
			this->pixel_metadata_cache.size = w * h;
		}

		this->pixel_trace_elements_allocator.reset_allocations();
		VoxelPainterImpl painter{w, h, this->pixel_metadata_cache.buffer, &this->pixel_trace_elements_allocator};

		for (std::int_fast64_t i = w * h - 1; i >= 0; i--) {
			new(&this->pixel_metadata_cache.buffer[i]) pixel_trace_metadata{};
		}

		for (; sprites_count-- > 0; sprites++) {
			painter.current_artist = *sprites;
			(*sprites)->draw(&painter, sprite_data);
		}

		std::uint_fast16_t y = 0;
		std::uint_fast16_t x = 0;

		for (std::int_fast64_t i = w * h - 1; i >= 0; i--) {
			this->pixel_metadata_cache.buffer[i].voxels = sort_zelements_desc(this->pixel_metadata_cache.buffer[i].voxels);
			RendererApi::Color cc = bg;
			RendererApi::SpriteMetadata *owner = nullptr;
			for (auto p = this->pixel_metadata_cache.buffer[i].voxels; p != nullptr; p = p->next) {
				cc = apply_transparent_color(cc, p->color);
				owner = p->owner;
			}
			frame->set_pixel(x, y, cc, owner);

			if (++x >= w) {
				y++;
				x = 0;
			}
		}
	}
}

#include "_default_renderer_linear_intrinsics.hpp"