#pragma once


#include <bgtu/computer_graphics_lab_work/renderer_api/color.hpp>
#include <bgtu/computer_graphics_lab_work/renderer_api/point.hpp>
#include "matrix.hpp"


namespace BGTU::ComputerGraphicsLabWork::Utilities {
	class Shader {
	public:
		[[nodiscard]] virtual RendererApi::Color::Transparent get_color(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) const = 0;

		[[nodiscard]]  inline RendererApi::Color::Transparent get_color(RendererApi::PointI<2> p) const {
			return this->get_color(p.x, p.y);
		};

		class Empty;

		static Shader::Empty const empty;
	};

	class Shader::Empty : public Shader {
	public:
		constexpr Empty() noexcept = default;

		[[nodiscard]] constexpr RendererApi::Color::Transparent get_color(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) const noexcept final {
			return RendererApi::Color::Transparent{0, 0, 0, 0};
		};

		operator Shader::Empty const *() const noexcept { // NOLINT(*-explicit-constructor)
			return this;
		}

		Shader::Empty const *operator->() const noexcept {
			return this;
		}

	};


	constexpr bool operator==(Shader::Empty const &, Shader::Empty const &) noexcept {
		return true;
	}

	constexpr bool operator==(Shader::Empty const &, Shader::Empty const *) noexcept {
		return true;
	}

	constexpr bool operator==(Shader::Empty const &, Shader const *other) noexcept {
		return dynamic_cast<Shader::Empty const *>(other) != nullptr;
	}

	constexpr bool operator!=(Shader::Empty const &, Shader::Empty const &) noexcept {
		return false;
	}

	constexpr bool operator!=(Shader::Empty const &, Shader::Empty const *) noexcept {
		return false;
	}

	constexpr bool operator!=(Shader::Empty const &, Shader const *other) noexcept {
		return dynamic_cast<Shader::Empty const *>(other) == nullptr;
	}

	constexpr bool operator==(Shader::Empty const *, Shader::Empty const &) noexcept {
		return true;
	}

	constexpr bool operator==(Shader const *other, Shader::Empty const &) noexcept {
		return dynamic_cast<Shader::Empty const *>(other) != nullptr;
	}

	constexpr bool operator!=(Shader::Empty const *, Shader::Empty const &) noexcept {
		return false;
	}

	constexpr bool operator!=(Shader const *other, Shader::Empty const &) noexcept {
		return dynamic_cast<Shader::Empty const *>(other) == nullptr;
	}

	constexpr Shader::Empty const Shader::empty{};

	class MonoShader : public Shader {
	private:
		RendererApi::Color::Transparent c;

	public:
		constexpr explicit MonoShader(RendererApi::Color::Transparent c) noexcept: c{c} {}

		[[nodiscard]] constexpr RendererApi::Color::Transparent get_color(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) const noexcept final {
			return this->c;
		};

		template<RendererApi::Color::Transparent _color>
		class Static : public Shader {
		public:
			constexpr Static() noexcept = default;

			static constexpr RendererApi::Color::Transparent const COLOR = _color;

			[[nodiscard]] constexpr RendererApi::Color::Transparent get_color(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) const noexcept final {
				return _color;
			};

		private:
			static MonoShader::Static<_color> const INSTANCE;

		public:
			operator MonoShader::Static<_color> const *() const noexcept { // NOLINT(*-explicit-constructor)
				return &INSTANCE;
			}


			MonoShader::Static<_color> const *operator->() const noexcept {
				return &INSTANCE;
			}


			template<RendererApi::Color::Transparent other_color>
			constexpr bool operator==(MonoShader::Static<other_color>) const noexcept {
				return _color == other_color;
			}
		};

		template<RendererApi::Color::Transparent color>
		static MonoShader::Static<color> const static_;
	};

	template<RendererApi::Color::Transparent color>
	constexpr MonoShader::Static<color> const MonoShader::Static<color>::INSTANCE{};

	template<RendererApi::Color::Transparent color>
	constexpr MonoShader::Static<color> const MonoShader::static_{};


	class TransformedShader : public Shader {
	private:
		Shader const *shader;
		Matrix4d transform;

	public:
		inline TransformedShader(Shader const *s, Matrix4d t) : shader{s}, transform{t} {}

		[[nodiscard]] inline RendererApi::Color::Transparent get_color(RendererApi::PointI<2>::component_t x, RendererApi::PointI<2>::component_t y) const override {
			RendererApi::PointF<3> in{x, y, 0};
			RendererApi::PointF<3> out = this->transform * in;
			return this->shader->get_color(out.x, out.y);
		}
	};
}